//
// UNIX SLang-based console implementation
// Copyright (c) 1996-9 Alex Ermolovich
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
#include "_united.h"
#include "console.h"
#include "Config.h"
#include <slang.h>

unsigned short int cMaxX, cMaxY;
static unsigned short int*vmem,a0,ox,ol,screenX=0,screenY=0;
static short int pushed_key;

#ifdef CON_XLATE_OUT
static unsigned char*_translate_out;
#endif
#ifdef CON_XLATE_IN
static unsigned char*_translate_in;
#endif

#if defined (CON_XLATE_OUT) || defined (CON_XLATE_IN)
#include "xlat.h"
#endif

void c_showcur(void) {SLtt_set_cursor_visibility(1);}
void c_hidecur(void) {SLtt_set_cursor_visibility(0);}
void c_movecur(int x,int y) {SLsmg_gotorc(y,x);screenY=y;screenX=x;SLsmg_refresh();}

void c_init(char*title)
{
 char**bg,**fg,*Colors[]={
 "black","blue","green","cyan","red","magenta","brown","lightgray",
 "gray","brightblue","brightgreen","brightcyan","brightred","brightmagenta","yellow","white",
 NULL}; 
 
 SLang_init_tty(0,0,0);
 SLtt_init_video();
 SLtt_get_terminfo();
 SLtt_get_screen_size();
 SLsmg_init_smg();
 SLkp_init();
 pushed_key=0;
 for(a0=0,bg=Colors;*bg;bg++) for(fg=Colors;*fg;fg++,a0++) SLtt_set_color(a0,NULL,*fg,*bg);
 cMaxY=SLtt_Screen_Rows-CON_ROW_FIX;
 cMaxX=SLtt_Screen_Cols-CON_COL_FIX; 
 vmem=(unsigned short int*)malloc(cMaxX*cMaxY*sizeof(unsigned short int));
 SLsmg_cls();
 c_hidecur();
#ifdef CON_XLATE_OUT
 _translate_out=(unsigned char*)getenv("RTN_CON_XLATE_OUT");
 if(_translate_out)
  _translate_out=_xlat_char[atoi((char*)_translate_out)];
 else
  _translate_out=_xlat_char[CON_XLATE_OUT];
#endif
#ifdef CON_XLATE_IN
 _translate_in=(unsigned char*)getenv("RTN_CON_XLATE_IN");
 if(_translate_in)
  _translate_in=_xlat_char[atoi((char*)_translate_in)];
 else
  _translate_in=_xlat_char[CON_XLATE_IN];
#endif
}

void c_settitle(char*title){}

void c_done(void)
{
 free(vmem);
 c_showcur();
 SLsmg_cls();
 SLsmg_refresh();
 SLsmg_reset_smg();
 SLtt_reset_video();
 SLang_reset_tty(); 
}

#define CHAR_PART 0x000ff
#define ATTR_PART 0x0ff00

static void repaint(short int x,short int y,short int len)
{
 SLsmg_gotorc(y,x);
 SLsmg_write_raw(vmem+y*cMaxX+x,len);
 SLsmg_gotorc(screenY,screenX);
 SLsmg_refresh();
}

void c_putc(const char*buf,short int len,short int x,short int y)
{
 for(ox=x,ol=len;len;len--,x++)
 {
  vmem[y*cMaxX+x]&=ATTR_PART;
  vmem[y*cMaxX+x]|=
#ifdef CON_XLATE_OUT
  _translate_out[(unsigned char)*buf++];
#else
  (unsigned char)*buf++;
#endif
 }
 repaint(ox,y,ol);
}

void c_puta(const char*buf,short int len,short int x,short int y)
{
 for(ox=x,ol=len;len;len--,x++) {vmem[y*cMaxX+x]&=CHAR_PART;vmem[y*cMaxX+x]|=(127&*buf++)<<8;}
 repaint(ox,y,ol);
}

void c_fillc(const char buf,short int len,short int x,short int y)
{
 for(ox=x,ol=len;len;len--,x++)
 {
  vmem[y*cMaxX+x]&=ATTR_PART;
  vmem[y*cMaxX+x]|=
#ifdef CON_XLATE_OUT
  _translate_out[(unsigned char)buf];
#else
  (unsigned char)buf;
#endif
 }
 repaint(ox,y,ol);
}

void c_filla(const char buf,short int len,short int x,short int y)
{
 for(ox=x,ol=len;len;len--,x++) {vmem[y*cMaxX+x]&=CHAR_PART;vmem[y*cMaxX+x]|=(buf&127)<<8;}
 repaint(ox,y,ol);
}

void c_writeca(const char*buf,const char abuf,short int len,short int x,short int y)
{
 for(a0=(abuf&127)<<8,ox=x,ol=len;len;len--,x++)
 {
  vmem[y*cMaxX+x]=a0+
#ifdef CON_XLATE_OUT
  _translate_out[(unsigned char)*buf++];
#else
  (unsigned char)*buf++;
#endif
 }
 repaint(ox,y,ol);
}

void c_scrollca(char dir,short int x,short int y,short int w,short int h)
{
 h--;
 if(dir==SCROLL_UP)
  for(int z=0;z<h;z++)
  {
   memcpy(vmem+(y+z)*cMaxX+x,vmem+(1+y+z)*cMaxX+x,w<<1);
   repaint(x,y+z,w);
  }
 else
  for(--h;h>=0;h--)
  {
   memcpy(vmem+(1+y+h)*cMaxX+x,vmem+(y+h)*cMaxX+x,w<<1);
   repaint(x,1+y+h,w);
  }
}

void c_pushkey(short int newkey)
{
 pushed_key=newkey;
}

short int c_getkey(void)
{
 if(pushed_key)
 {
  a0=pushed_key;
  pushed_key=0;
  return a0;
 }
 a0=SLang_getkey();
 if(a0==033) if(!SLang_input_pending(0)) return _KBD_ESC;
 SLang_ungetkey(a0);
 a0=SLkp_getkey();
 switch(a0)
 {
  case SL_KEY_BACKSPACE:a0=_KBD_BS;break;
  case SL_KEY_F(1)     :a0=_KBD_F1;break;
  case SL_KEY_F(2)     :a0=_KBD_F2;break;
  case SL_KEY_F(3)     :a0=_KBD_F3;break;
  case SL_KEY_F(4)     :a0=_KBD_F4;break;
  case SL_KEY_F(5)     :a0=_KBD_F5;break;
  case SL_KEY_F(6)     :a0=_KBD_F6;break;
  case SL_KEY_F(7)     :a0=_KBD_F7;break;
  case SL_KEY_F(8)     :a0=_KBD_F8;break;
  case SL_KEY_F(9)     :a0=_KBD_F9;break;
  case SL_KEY_UP       :a0=_KBD_UP;break;
  case SL_KEY_F(10)    :a0=_KBD_F10;break;
  case SL_KEY_DELETE   :a0=_KBD_DEL;break;
  case SL_KEY_IC       :a0=_KBD_INS;break;  
  case SL_KEY_END      :a0=_KBD_END;break;
  case SL_KEY_DOWN     :a0=_KBD_DOWN;break;
  case SL_KEY_LEFT     :a0=_KBD_LEFT;break;
  case SL_KEY_PPAGE    :a0=_KBD_PGUP;break;
  case SL_KEY_NPAGE    :a0=_KBD_PGDN;break;
  case SL_KEY_HOME     :a0=_KBD_HOME;break;
  case SL_KEY_ENTER    :a0=_KBD_ENTER;break;
  case SL_KEY_RIGHT    :a0=_KBD_RIGHT;break;
  default:
#ifdef CON_XLATE_IN
  a0=_translate_in[(unsigned)a0];
#endif  
 }
 return a0;
}
