//
// FTN Type2+ packet handling class
// Copyright (c) 1996-9 Alex Ermolovich
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
#include "ftnpkt.h"
#include "_united.h"

#define BULLETPROOF
// improves vitality, decreases speed

char FTNPacket::IncCurrent(void)
{
 if(_tempo_msg->next)
 {
  _tempo_msg=_tempo_msg->next;
  _current_msg=_tempo_msg->msg;
  _current_msg_num++;
  if(_current_msg==_replica) reload();
  return 1;
 }
 else
  return 0;
}

char FTNPacket::DecCurrent(void)
{
 if(_tempo_msg->prev)
 {
  _tempo_msg=_tempo_msg->prev;
  _current_msg=_tempo_msg->msg;
  _current_msg_num--;
  if(_current_msg==_replica) reload();
  return 1;
 }
 else
  return 0;
}

void FTNPacket::DeleteCurrent(void)
{
 if(!_tempo_msg) return;
 if(_current_msg!=_replica) delete _tempo_msg->msg;
 _total_msg_num--;
 _tempo_msg_0=_tempo_msg;
 if(_tempo_msg==_first_msg)
 {
  _first_msg=_tempo_msg->next;
  _tempo_msg=_first_msg;
  if(_tempo_msg) _tempo_msg->prev=0;
 }
 else
  if(_tempo_msg->next)
  {
   _tempo_msg->prev->next=_tempo_msg->next;
   _tempo_msg->next->prev=_tempo_msg->prev;
   _tempo_msg=_tempo_msg->next;
  }
  else
  {
   _tempo_msg->prev->next=NULL;
   _tempo_msg=_tempo_msg->prev;
   _current_msg_num--;
  }
 delete _tempo_msg_0;
 if(_tempo_msg)
 {
  _current_msg=_tempo_msg->msg;
  if(_current_msg==_replica) reload();
 }
 else
  _current_msg=NULL;
}

void FTNPacket::Insert(FTNMessage*msg)
{
 state=pktOK;
 if(!_first_msg)
 {
  if(!(_first_msg=new FTNMsgList)) {state=pktNOMEM;return;}
  _tempo_msg=_first_msg;
 }
 else
 {
  _tempo_msg=_first_msg;
  while(_tempo_msg->next) _tempo_msg=_tempo_msg->next;
  if(!(_tempo_msg->next=new FTNMsgList)) {state=pktNOMEM;return;}
  _tempo_msg=_tempo_msg->next;
 }
 _tempo_msg->next=0;
 _tempo_msg->msg=msg;
 _current_msg_num=_total_msg_num++;
 _current_msg=_tempo_msg->msg;
}

void FTNPacket::_boot(void)
{
 state=pktOK;
 _replica=_current_msg=0;
 _tempo_msg=_first_msg=0;
 _total_msg_num=_current_msg_num=0;
 if(!(_tempo_file=new fstream)) state=pktNOMEM;
}

FTNPacket::FTNPacket()
{
 _boot();
}

FTNPacket::FTNPacket(FTNPacket*tpl)
{
 _boot();
 _CW                 =tpl->_CW;
 _day                =tpl->_day;
 _hour               =tpl->_hour;
 _baud               =tpl->_baud;
 _year               =tpl->_year;
 _month              =tpl->_month;
 _AuxNet             =tpl->_AuxNet;
 _minute             =tpl->_minute;
 _second             =tpl->_second;
 _orignet            =tpl->_orignet;
 _destnet            =tpl->_destnet;
 _orignode           =tpl->_orignode;
 _destnode           =tpl->_destnode;
 _origZone           =tpl->_origZone;
 _destZone           =tpl->_destZone;
 _prodrev0           =tpl->_prodrev0;
 _prodrev1           =tpl->_prodrev1;
 _origPoint          =tpl->_origPoint;
 _destPoint          =tpl->_destPoint;
 memcpy(psw          ,tpl->psw,8);
 memcpy(psd          ,tpl->psd,4);
 strcpy(_to_address  ,tpl->_to_address);
 strcpy(_from_address,tpl->_from_address);
}

FTNPacket::~FTNPacket()
{
 delete _tempo_file;
}

void FTNPacket::Dispell(void)
{
 _tempo_msg=_first_msg;
 while(_tempo_msg)
 {
  if(_tempo_msg->msg!=_replica) delete _tempo_msg->msg;
  _tempo_msg_0=_tempo_msg;
  _tempo_msg=_tempo_msg->next;
  delete _tempo_msg_0;
 }
 if(_replica) delete _replica;
 _first_msg=0;
}

void FTNPacket::DirtyDispell(void)
{
 _tempo_msg=_first_msg;
 while(_tempo_msg)
 {
  _tempo_msg_0=_tempo_msg;
  _tempo_msg=_tempo_msg->next;
  delete _tempo_msg_0;
 }
 if(_replica) delete _replica;
}

void FTNPacket::reload(void)
{
 state=pktOK;
 _tempo_file->seekg(_tempo_msg->seekpos);
 _tempo_msg->msg->ReadPacked(_tempo_file);
 if(_tempo_msg->msg->state!=msgOK) state=_tempo_msg->msg->state;
}

char FTNPacket::SeekCurrent(short int _selected)
{
 if(_selected>=_total_msg_num) return 0;
 if(_selected==_current_msg_num) return 1;
 if(_current_msg_num<_selected)
  while(_current_msg_num!=_selected)
  {
   _tempo_msg=_tempo_msg->next;
   _current_msg=_tempo_msg->msg;
   _current_msg_num++;
  }
 else
  while(_current_msg_num!=_selected)
  {
   _tempo_msg=_tempo_msg->prev;
   _current_msg=_tempo_msg->msg;
   _current_msg_num--;
  }
 if(_current_msg==_replica) reload();
 return 1;
}

char FTNPacket::_read_init(char*name)
{
 state=pktOK;
 _first_msg=0;
 _total_msg_num=_current_msg_num=0;
 if(access(name,0)==-1) {state=pktBADPKT;return 0;}
 _tempo_file->open(name,ios::in|ios::fakebinary);
 _tempo_file->read((char*)&_orignode,2);
 _tempo_file->read((char*)&_destnode,2);
 _tempo_file->read((char*)&_year,2);
 _tempo_file->read((char*)&_month,2);
 _tempo_file->read((char*)&_day,2);
 _tempo_file->read((char*)&_hour,2);
 _tempo_file->read((char*)&_minute,2);
 _tempo_file->read((char*)&_second,2);
 _tempo_file->read((char*)&_baud,2);
 _tempo_file->read((char*)&_tempo_int,2);
 if(_tempo_int!=2) {state=pktBADPKT;return 0;}
 _tempo_file->read((char*)&_orignet,2);
 _tempo_file->read((char*)&_destnet,2);
 _tempo_file->read((char*)&_prodrev0,2);
 _tempo_file->read(psw,8);
 _tempo_file->read((char*)&_origZone,2);
 _tempo_file->read((char*)&_destZone,2);
 _tempo_file->read((char*)&_AuxNet,2);
 _tempo_file->read((char*)&_tempo_int,2);
 _tempo_file->read((char*)&_prodrev1,2);
 _tempo_file->read((char*)&_CW,2);
 if(_tempo_int!=0x100*(char)_CW+_CW/0x100) {state=pktBADPKT;return 0;}
 _tempo_file->read((char*)&_origZone,2);
 _tempo_file->read((char*)&_destZone,2);
 _tempo_file->read((char*)&_origPoint,2);
 _tempo_file->read((char*)&_destPoint,2);
 sprintf(_to_address,"%d:%d/%d",_destZone,_destnet,_destnode);
 if(_destPoint) sprintf(_to_address+strlen(strcat(_to_address,".")),"%d",_destPoint);
 sprintf(_from_address,"%d:%d/%d",_origZone,_orignet,_orignode);
 if(_origPoint) sprintf(_from_address+strlen(strcat(_from_address,".")),"%d",_origPoint);
 _tempo_file->read(psd,4);
 if(!(_tempo_msg_0=new FTNMsgList)) {state=pktNOMEM;return 0;}
 if(!(_tempo_msg_0->msg=new FTNMessage)) {state=pktNOMEM;return 0;}
 return 1;
}

char FTNPacket::ReadFlat(char*name)
{
 if(!_read_init(name)) return 0;
 while (_tempo_msg_0->msg->ReadPacked(_tempo_file))
 {
  if(_tempo_msg_0->msg->state!=msgOK) {state=_tempo_msg_0->msg->state;return 0;}
  if(!_first_msg)
  {
   _first_msg=_tempo_msg_0;
   _tempo_msg=_first_msg;
   _tempo_msg->prev=_tempo_msg->next=NULL;
  }
  else
  {
   _tempo_msg->next=_tempo_msg_0;
   _tempo_msg_0->prev=_tempo_msg;
   _tempo_msg=_tempo_msg_0;
   _tempo_msg->next=NULL;
  }
  if(!(_tempo_msg_0=new FTNMsgList)) {state=pktNOMEM;return 0;}
  if(!(_tempo_msg_0->msg=new FTNMessage)) {state=pktNOMEM;return 0;}
  _total_msg_num++;
 }
 delete _tempo_msg_0->msg;
 delete _tempo_msg_0;
 _tempo_msg=_first_msg;
 if(_tempo_msg) _current_msg=_tempo_msg->msg; else _current_msg=NULL;
 _tempo_file->close();
 return 1;
}

char FTNPacket::ReadLookup(char*name)
{
 char oldstate;

 if(!_read_init(name)) return 0;
 _tempo_msg_0->seekpos=_tempo_file->tellg();
 while (seektonext())
 {
  if(!_first_msg)
  {
   _first_msg=_tempo_msg_0;
   _tempo_msg=_first_msg;
   _tempo_msg->prev=_tempo_msg->next=NULL;
  }
  else
  {
   _tempo_msg->next=_tempo_msg_0;
   _tempo_msg_0->prev=_tempo_msg;
   _tempo_msg=_tempo_msg_0;
   _tempo_msg->next=NULL;
  }
  if(!(_tempo_msg_0=new FTNMsgList)) {state=pktNOMEM;return 0;}
  _tempo_msg_0->msg=_first_msg->msg;
  _tempo_msg_0->seekpos=_tempo_file->tellg();
  _total_msg_num++;
 }
 delete _tempo_msg_0;
 _tempo_msg=_first_msg;
 if(_tempo_msg)
 {
  _replica=_current_msg=_tempo_msg->msg;
  oldstate=state;
  reload();
  state|=oldstate;
 }
 else
  _current_msg=NULL;
 return 1;
}

char FTNPacket::seektonext(void)
{
 char tempo;
 if(_tempo_file->peek()==EOF) {_tempo_file->clear();state=pktUNEXPECTED_EOF; return 0;}
 _tempo_file->read((char*)&_tempo_int,1);
 if(_tempo_file->peek()==EOF) {_tempo_file->clear();state=pktUNEXPECTED_EOF; return 0;}
 _tempo_file->read(1+(char*)&_tempo_int,1);
 if(_tempo_int!=2)
  if(_tempo_file->peek()==EOF)
  {_tempo_file->clear();state=pktOK; return 0;}
  else
  {_tempo_file->clear();state=pktMISSING_EOF; return 0;}
 _tempo_file->seekg(_tempo_file->tellg()+12);
 _tempo_file->read(_tempo_msg_0->datetime,20);
#ifdef BULLETPROOF
#define READSTRING(A,B) {_tempo_str_ptr=A;_tempo_int=0;do {if(_tempo_int++==B){*--_tempo_str_ptr=0;break;}_tempo_file->read(_tempo_str_ptr,1);} while(*_tempo_str_ptr++);}
 READSTRING(_tempo_msg_0->to,36)
 READSTRING(_tempo_msg_0->from,36)
 READSTRING(_tempo_msg_0->subj,72)
#undef READSTRING
#else
#define READSTRING(A) {_tempo_str_ptr=A; do _tempo_file->read(_tempo_str_ptr,1); while(*_tempo_str_ptr++);}
 READSTRING(_tempo_msg_0->to)
 READSTRING(_tempo_msg_0->from)
 READSTRING(_tempo_msg_0->subj)
#undef READSTRING
#endif
 *_tempo_msg_0->area=0;
 *_tempo_file>>tempo;
 if(tempo=='A')
 {
  *_tempo_file>>tempo;
  if(tempo=='R')
  {
   *_tempo_file>>tempo;
   if(tempo=='E')
   {
    *_tempo_file>>tempo;
    if(tempo=='A')
    {
     *_tempo_file>>tempo;
     if(tempo==':')
     {
      _tempo_str_ptr_0=_tempo_msg_0->area;
#ifdef BULLETPROOF
      _tempo_int=0;do {if(_tempo_int++==61)break;_tempo_file->read(_tempo_str_ptr_0,1);} while(*_tempo_str_ptr_0++!=0x0d);
#else
      do _tempo_file->read(_tempo_str_ptr_0,1); while(*_tempo_str_ptr_0++!=0x0d);
#endif
      *--_tempo_str_ptr_0=0;
     }
    }
   }
  }
 }
 while (tempo&&_tempo_file->peek()!=EOF) *_tempo_file>>tempo;
 if(_tempo_file->peek()==EOF) {_tempo_file->clear();state=pktUNEXPECTED_EOF; return 0;}
 return 1;
}

void FTNPacket::Write(char*name)
{
 char *xname=NULL;

 state=pktOK;
 if(!_tempo_msg)
 {
  if(_replica) _tempo_file->close();
  remove(name);
 }
 else
 {
  if(_replica)
  {
   if(!(xname=strdup(name))) {state=pktNOMEM;return;}
   xname[strlen(xname)-1]='%';
   if(!(_tempo_file_0=new fstream)) {state=pktNOMEM;return;}
   _tempo_file_0->open(xname,ios::out|ios::fakebinary|ios::trunc);
  }
  else
  {
   _tempo_file_0=_tempo_file;
   _tempo_file_0->open(name,ios::out|ios::fakebinary|ios::trunc);
  }
  _tempo_file_0->write((char*)&_orignode,2);
  _tempo_file_0->write((char*)&_destnode,2);
  _tempo_file_0->write((char*)&_year,2);
  _tempo_file_0->write((char*)&_month,2);
  _tempo_file_0->write((char*)&_day,2);
  _tempo_file_0->write((char*)&_hour,2);
  _tempo_file_0->write((char*)&_minute,2);
  _tempo_file_0->write((char*)&_second,2);
  _tempo_file_0->write((char*)&_baud,2);
  _tempo_int=2;
  _tempo_file_0->write((char*)&_tempo_int,2);
  _tempo_file_0->write((char*)&_orignet,2);
  _tempo_file_0->write((char*)&_destnet,2);
  _tempo_file_0->write((char*)&_prodrev0,2);
  _tempo_file_0->write(psw,8);
  _tempo_file_0->write((char*)&_origZone,2);
  _tempo_file_0->write((char*)&_destZone,2);
  _tempo_file_0->write((char*)&_AuxNet,2);
  _tempo_int=0x100*(char)_CW+_CW/0x100;
  _tempo_file_0->write((char*)&_tempo_int,2);
  _tempo_file_0->write((char*)&_prodrev1,2);
  _tempo_file_0->write((char*)&_CW,2);
  _tempo_file_0->write((char*)&_origZone,2);
  _tempo_file_0->write((char*)&_destZone,2);
  _tempo_file_0->write((char*)&_origPoint,2);
  _tempo_file_0->write((char*)&_destPoint,2);
  _tempo_file_0->write(psd,4);
  _tempo_msg=_first_msg;
  while(_tempo_msg)
  {
   if(_tempo_msg->msg==_replica) reload();
   _tempo_msg->msg->WritePacked(_tempo_file_0);
   _tempo_msg=_tempo_msg->next;
  }
  _tempo_int=0;
  _tempo_file_0->write((char*)&_tempo_int,2);
  _tempo_file_0->close();
  if(_replica)
  {
   delete _tempo_file_0;
   _tempo_file->close();
   remove(name);
   rename(xname,name);
   free(xname);
  }
 }
}

void FTNPacket::WriteChunk(void)
{
 if(_current_msg)
 {
  _current_msg->WritePacked(_tempo_file);
  _current_msg=0;
 }
}

void FTNPacket::WriteChunkFixup(void)
{
 if(_current_msg)
 {
  _current_msg->_destnode=_destnode;
  _current_msg->_destnet=_destnet;
  _current_msg->WritePacked(_tempo_file);
  _current_msg=0;
 }
}

void FTNPacket::WriteChunkInit(char*name)
{
 state=pktOK;
 _tempo_file->open(name,ios::out|ios::fakebinary|ios::trunc);
 _tempo_file->write((char*)&_orignode,2);
 _tempo_file->write((char*)&_destnode,2);
 _tempo_file->write((char*)&_year,2);
 _tempo_file->write((char*)&_month,2);
 _tempo_file->write((char*)&_day,2);
 _tempo_file->write((char*)&_hour,2);
 _tempo_file->write((char*)&_minute,2);
 _tempo_file->write((char*)&_second,2);
 _tempo_file->write((char*)&_baud,2);
 _tempo_int=2;
 _tempo_file->write((char*)&_tempo_int,2);
 _tempo_file->write((char*)&_orignet,2);
 _tempo_file->write((char*)&_destnet,2);
 _tempo_file->write((char*)&_prodrev0,2);
 _tempo_file->write(psw,8);
 _tempo_file->write((char*)&_origZone,2);
 _tempo_file->write((char*)&_destZone,2);
 _tempo_file->write((char*)&_AuxNet,2);
 _tempo_int=0x100*(char)_CW+_CW/0x100;
 _tempo_file->write((char*)&_tempo_int,2);
 _tempo_file->write((char*)&_prodrev1,2);
 _tempo_file->write((char*)&_CW,2);
 _tempo_file->write((char*)&_origZone,2);
 _tempo_file->write((char*)&_destZone,2);
 _tempo_file->write((char*)&_origPoint,2);
 _tempo_file->write((char*)&_destPoint,2);
 _tempo_file->write(psd,4);
}

void FTNPacket::WriteChunkDone(void)
{
 _tempo_int=0;
 _tempo_file->write((char*)&_tempo_int,2);
 _tempo_file->close();
}

void FTNPacket::InsertChunk(FTNMessage*msg)
{
 _current_msg=msg;
 _total_msg_num++;
}
