//
// UNIX X11-based console implementation
// Copyright (c) 1996-9 Alex Ermolovich
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
#include "_united.h"
#include "console.h"
#include "Config.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

unsigned short int cMaxX, cMaxY;

static GC gccache[16*16];
static Window win;
static Display*disp;
static XEvent event;
static Colormap cmap;
static unsigned char *vmem_a,*vmem_c,*cache_a,*cache_c,runner,cursor=0;
static unsigned short int zlen,screenX=0,screenY=0,fnt_width,fnt_height;
static short int pushed_key,facelift;

#ifdef X11CON_XLATE_OUT
static unsigned char*_translate_out;
#endif
#ifdef X11CON_XLATE_IN
static unsigned char*_translate_in;
#endif

#if defined (X11CON_XLATE_OUT) || defined (X11CON_XLATE_IN)
#include "xlat.h"
#endif

#include "16colors.h"
#include "keytrans.h"

inline void cur_show(void) {XDrawLine(disp,win,gccache[vmem_a[screenY*cMaxX+screenX]],screenX*fnt_width,(screenY+1)*fnt_height-facelift,(screenX+1)*fnt_width-1,(screenY+1)*fnt_height-facelift);}
inline void cur_hide(void) {XDrawImageString(disp,win,gccache[*(vmem_a+screenY*cMaxX+screenX)],screenX*fnt_width,(screenY+1)*fnt_height-facelift,(char*)(vmem_c+screenY*cMaxX+screenX),1);}

void c_showcur(void) {if(!cursor) {cursor=1;cur_show();}}
void c_hidecur(void) {if(cursor) {cur_hide();cursor=0;}}
void c_movecur(int x,int y) {if(cursor)cur_hide();screenY=y;screenX=x;if(cursor)cur_show();}

void c_init(char*title)
{
 XColor cmapen;
 XSizeHints xsh;
 XGCValues data;
 XFontStruct*fnt;
 unsigned long colors_data[16];
 XWMHints xwmh={InputHint|StateHint,True,NormalState,0,0,0,0,0,0};

 pushed_key=0;
 disp=XOpenDisplay(NULL);
 if(getenv("RTN_X11CON_SIZE"))
 {
  cMaxX=atoi(getenv("RTN_X11CON_SIZE"));
  cMaxY=atoi(1+strchr(getenv("RTN_X11CON_SIZE"),'x'));
 } 
 else
 {
  cMaxX=X11CON_WIDTH;
  cMaxY=X11CON_HEIGHT;
 }
 if(getenv("RTN_X11CON_FONT"))
  fnt=XLoadQueryFont(disp,getenv("RTN_X11CON_FONT"));
 else
  fnt=XLoadQueryFont(disp,X11CON_FONT);
 fnt_width=fnt->max_bounds.width;
 fnt_height=fnt->max_bounds.ascent+(facelift=fnt->max_bounds.descent);
 xsh.flags=PMaxSize|PMinSize;
 win=XCreateSimpleWindow(disp,DefaultRootWindow(disp),0,0,xsh.max_width=xsh.min_width=cMaxX*fnt_width,xsh.max_height=xsh.min_height=cMaxY*fnt_height,0,0,0);
 XSetStandardProperties(disp,win,title,title,None,0,0,&xsh);
 XSetWMHints(disp,win,&xwmh);
 XSelectInput(disp,win,ExposureMask|KeyReleaseMask|KeymapStateMask);
 XClearWindow(disp,win);
 cmap=XCreateColormap(disp,win,DefaultVisual(disp,DefaultScreen(disp)),AllocNone);
 for(runner=0;runner<16;runner++)
 {
  cmapen.red  =colors_rgb[runner*3];
  cmapen.green=colors_rgb[runner*3+1];
  cmapen.blue =colors_rgb[runner*3+2];
  XAllocColor(disp,cmap,&cmapen);
  colors_data[runner]=cmapen.pixel;
 }
 XInstallColormap(disp,cmap);
 XSetWindowColormap(disp,win,cmap);
 XMapWindow(disp, win);
 vmem_a=new unsigned char[cMaxX*cMaxY];
 vmem_c=new unsigned char[cMaxX*cMaxY];
 data.fill_style=FillSolid;
 data.font=fnt->fid;
 for(zlen=0;zlen<16*16;zlen++)
 {
  data.foreground=colors_data[zlen%16];
  data.background=colors_data[zlen/16];
  gccache[zlen]=XCreateGC(disp,win,GCFont|GCFillStyle|GCForeground|GCBackground,&data);
 }
#ifdef X11CON_XLATE_OUT
 _translate_out=(unsigned char*)getenv("RTN_X11CON_XLATE_OUT");
 if(_translate_out)
  _translate_out=_xlat_char[atoi((char*)_translate_out)];
 else
  _translate_out=_xlat_char[X11CON_XLATE_OUT];
#endif
#ifdef X11CON_XLATE_IN
 _translate_in=(unsigned char*)getenv("RTN_X11CON_XLATE_IN");
 if(_translate_in)
  _translate_in=_xlat_char[atoi((char*)_translate_in)];
 else
  _translate_in=_xlat_char[X11CON_XLATE_IN];
#endif
}

void c_settitle(char*title)
{
 XSetStandardProperties(disp,win,title,title,None,0,0,NULL);
}

void c_done(void)
{
 XUninstallColormap(disp,cmap);
 XCloseDisplay(disp);
 delete[] vmem_a;
 delete[] vmem_c;
}

inline void repaint(short int x,short int y,short int len)
{
 int oldx=x,oldy=y,oldlen=len;
#ifdef NO_ACCEL
 for(cache_a=vmem_a+y*cMaxX+x,cache_c=vmem_c+y*cMaxX+x,y=(y+1)*fnt_height-facelift,x*=fnt_width;len--;x+=fnt_width,cache_c++,cache_a++)
  XDrawImageString(disp,win,gccache[*cache_a],x,y,cache_c,1);
#else
 for(cache_a=vmem_a+y*cMaxX+x,cache_c=vmem_c+y*cMaxX+x,y=(y+1)*fnt_height-facelift,x*=fnt_width;len;x+=fnt_width*zlen,cache_c+=zlen)
 {
  for(zlen=0,runner=*cache_a;len&&(runner==*cache_a);cache_a++,len--) zlen++;
  XDrawImageString(disp,win,gccache[runner],x,y,(char*)cache_c,zlen);
 }
#endif
 if(cursor&&(oldy==screenY)&&(oldx<=screenX)&&(oldx+oldlen>screenX)) cur_show();
}

void c_scrollca(char dir,short int x,short int y,short int w,short int h)
{
 h--;
 if(dir==SCROLL_UP)
  for(int z=0;z<h;z++)
  {
   memcpy(vmem_a+(y+z)*cMaxX+x,vmem_a+(1+y+z)*cMaxX+x,w);
   memcpy(vmem_c+(y+z)*cMaxX+x,vmem_c+(1+y+z)*cMaxX+x,w);
   repaint(x,y+z,w);
  }
 else
  for(--h;h>=0;h--)
  {
   memcpy(vmem_a+(1+y+h)*cMaxX+x,vmem_a+(y+h)*cMaxX+x,w);
   memcpy(vmem_c+(1+y+h)*cMaxX+x,vmem_c+(y+h)*cMaxX+x,w);
   repaint(x,1+y+h,w);
  }
}

inline void catch_expose(void)
{
 do
 {
  int zx=((XExposeEvent*)&event)->x/fnt_width,
      zy=((XExposeEvent*)&event)->y/fnt_height,
      zxw=((XExposeEvent*)&event)->width/fnt_width,
      zyh=((XExposeEvent*)&event)->height/fnt_height;
  while((zx+zxw)*fnt_width<((XExposeEvent*)&event)->x+((XExposeEvent*)&event)->width) zxw++;
  while((zy+zyh)*fnt_height<((XExposeEvent*)&event)->y+((XExposeEvent*)&event)->height) zyh++;
  for(int y=0;y<zyh;y++) repaint(zx,zy+y,zxw);
 }while(XCheckTypedEvent(disp,Expose,&event));
}

void c_putc(const char*buf,short int len,short int x,short int y)
{
 while(len&&*(vmem_c+y*cMaxX+x)==*(unsigned char*)buf) {len--;x++;buf++;}
#ifdef X11CON_XLATE_OUT
 for(int xlatez=0;xlatez<len;xlatez++) *(vmem_c+y*cMaxX+x+xlatez)=_translate_out[*(unsigned char*)(buf+xlatez)];
#else
 memcpy(vmem_c+y*cMaxX+x,buf,len);
#endif
 repaint(x,y,len);
}

void c_puta(const char*buf,short int len,short int x,short int y)
{
 while(len&&*(vmem_a+y*cMaxX+x)==*(unsigned char*)buf) {len--;x++;buf++;}
 memcpy(vmem_a+y*cMaxX+x,buf,len);
 repaint(x,y,len);
}

void c_fillc(const char buf,short int len,short int x,short int y)
{
 while(len&&*(vmem_c+y*cMaxX+x)==(unsigned char)buf) {len--;x++;}
#ifdef X11CON_XLATE_OUT
 for(int xlatez=0;xlatez<len;xlatez++) *(vmem_c+y*cMaxX+x+xlatez)=_translate_out[(unsigned char)buf];
#else
 memset(vmem_c+y*cMaxX+x,buf,len);
#endif
 repaint(x,y,len);
}

void c_filla(const char buf,short int len,short int x,short int y)
{
 while(len&&*(vmem_a+y*cMaxX+x)==(unsigned char)buf) {len--;x++;}
 memset(vmem_a+y*cMaxX+x,buf,len);
 XDrawImageString(disp,win,gccache[(unsigned char)buf],x*fnt_width,(y+1)*fnt_height-facelift,(char*)(vmem_c+y*cMaxX+x),len);
 if(cursor&&(y==screenY)&&(x<=screenX)&&(x+len>screenX)) cur_show();
}

void c_writeca(const char*buf,const char abuf,short int len,short int x,short int y)
{
 while(len&&(*(vmem_c+y*cMaxX+x)==*buf)&&(*(vmem_a+y*cMaxX+x)==abuf)) {len--;x++;buf++;}
#ifdef X11CON_XLATE_OUT
 for(int xlatez=0;xlatez<len;xlatez++) *(vmem_c+y*cMaxX+x+xlatez)=_translate_out[*(unsigned char*)(buf+xlatez)];
#else
 memcpy(vmem_c+y*cMaxX+x,buf,len);
#endif
 memset(vmem_a+y*cMaxX+x,abuf,len);
 XDrawImageString(disp,win,gccache[(unsigned)abuf],x*fnt_width,(y+1)*fnt_height-facelift,(char*)(vmem_c+y*cMaxX+x),len);
 if(cursor&&(y==screenY)&&(x<=screenX)&&(x+len>screenX)) cur_show();
}

#define KEY_MODS_NUM 7
static char*KEY_MODS[KEY_MODS_NUM]={"Caps_Lock","Alt_L","Alt_R","Control_L","Control_R","Shift_L","Shift_R"};

void c_pushkey(short int newkey)
{
 pushed_key=newkey;
}

short int c_getkey(void)
{
 int mods=0;
 XKeyEvent*keyev;
 char*decoded=NULL;

 if(pushed_key)
 {
  zlen=pushed_key;
  pushed_key=0;
  return zlen;
 }
 do{
    do{
       XNextEvent(disp,&event);
       if(event.type==MappingNotify) XRefreshKeyboardMapping((XMappingEvent*)&event); // kbd mapping changes
       else if(event.type==Expose) catch_expose(); // exposures
      }while(event.type!=KeyRelease);
    decoded=XKeysymToString(XLookupKeysym(keyev=(XKeyEvent*)&event,0));
    if(decoded) for(runner=0;runner<KEY_MODS_NUM;runner++) if(!strcmp(decoded,KEY_MODS[runner])) {decoded=0;break;} // skip modifiers
   }while(!decoded);

 if((keyev->state)&4) mods|=_KBD_CTRL;
 if((keyev->state)&8) mods|=_KBD_ALT;

// cut out specials
 for(runner=0;runner<KEY_TRANS_NUM;runner++)
  if(!strcmp(decoded,KEY_TRANS[runner].name))
  {
   if((keyev->state)&1) mods|=_KBD_SHIFT;
   if((keyev->state)&2) mods|=_KBD_CAPS;
   return mods|KEY_TRANS[runner].value;
  }
 
// the normal ones
 XLookupString(keyev,(char*)&runner,1,0,0);
#ifdef X11CON_XLATE_IN
 return mods|_translate_in[runner];
#else
 return mods|runner;
#endif
}
